package model.fish.behavior;

import model.ILambda;
import model.RandNumGenerator;
import sysModel.fish.Fish;
import sysModel.fish.IFishBehavior;

/**
 * A fish that moves like a knight in chess.
 * <p/>
 * _ _ _ _ _ |_|0|_|1|_| |7|_|_|_|2| |_|_|X|_|_| |6|_|_|_|3| |_|5|_|4|_|
 * <p/>
 * If the fish is located at the X, then the locations labelled 0 through 7 should be reachable. The target location is
 * selected at random. When moving, the fish should always make two steps in one direction, then one step over, NOT one
 * step in one direction, then two steps over. While the destination is the same, the locations visited are different.
 * If the fish is blocked on the way to the target location, the fish should just stop in that place. After the move has
 * been executed, the fish is left facing in the direction of the last step (the "one over" step).
 *
 * @author Mathias Ricken
 */
public class KnightBehavior implements IFishBehavior {
    /**
     * Carry out movement behavior for the host fish.
     *
     * @param host fish to move.
     */
    public void move(final Fish host) {
        int target = RandNumGenerator.instance().nextInt(8);
        int turns = target / 2; // number of 90 degree turns
        host.turnRight(turns * Math.PI / 2.0);
        final boolean oneOverToLeft = (0 == (target % 2)); // true if the "one over" step goes to the left, false if to the right

        // try to make the first step forward
        host.tryMoveFwd(new ILambda() {
            public Object apply(Object param) {
                // if blocked, do nothing
                return null;
            }
        }, new ILambda() {
            public Object apply(Object param) {
                // if open, move
                ((ILambda)param).apply(null);

                // try to make the second step forward
                return host.tryMoveFwd(new ILambda() {
                    public Object apply(Object param) {
                        // if blocked, do nothing
                        return null;
                    }
                }, new ILambda() {
                    public Object apply(Object param) {
                        // if open, move
                        ((ILambda)param).apply(null);

                        // if over to the left...
                        if (oneOverToLeft) {
                            // ...turn left
                            host.turnLeft();
                        }
                        else {
                            // ...else turn right
                            host.turnRight();
                        }

                        // try to make the step over
                        return host.tryMoveFwd(new ILambda() {
                            public Object apply(Object param) {
                                // if blocked, do nothing
                                return null;
                            }
                        }, new ILambda() {
                            public Object apply(Object param) {
                                // if open, move
                                ((ILambda)param).apply(null);

                                return null;
                            }
                        });
                    }
                });
            }
        });
    }

    /**
     * Private singleton ctor.
     */
    private KnightBehavior() {
    }

    /**
     * Singleton instance.
     */
    public static final KnightBehavior Singleton = new KnightBehavior();

}
